# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

import json
import os
import shutil

import tensorstore as ts
import zarr
from tifffile import imread

SUPPORTED_DRIVERS = ["neuroglancer_precomputed", "n5", "zarr"]


# -- os utils ---
def listdir(path, ext=None):
    """
    Lists all files in the directory at "path". If an extension "ext" is
    provided, then only files containing "ext" are returned.

    Parameters
    ----------
    path : str
        Path to directory to be searched.

    ext : str, optional
       Extension of file type of interest. The default is None.

    Returns
    -------
    list
        List of all files in directory at "path" with extension "ext" if
        provided. Otherwise, list of all files in directory.

    """
    if ext is None:
        return [f for f in os.listdir(path)]
    else:
        return [f for f in os.listdir(path) if ext in f]


def mkdir(path):
    """
    Makes a directory at "path" if it does not already exist.

    Parameters
    ----------
    path : str
        Path to directory.

    Returns
    -------
    None.

    """
    if not os.path.exists(path):
        os.mkdir(path)


def rmdir(path):
    """
    Removes a directory at "path" if it already exist.

    Parameters
    ----------
    path : str
        Path to directory.

    Returns
    -------
    None.

    """
    if os.path.exists(path):
        shutil.rmtree(path)


# --- data structure utils ---
def check_edge(edge_list, edge):
    """
    Checks if "edge" is in "edge_list".

    Parameters
    ----------
    edge_list : list or set
        List or set of edges.
    edge : tuple
        Edge.

    Returns
    -------
    bool : bool
        Indication of whether "edge" is contained in "edge_list".

    """
    if edge in edge_list or (edge[1], edge[0]) in edge_list:
        return True
    else:
        return False


def remove_edge(edge_list, edge):
    """
    Checks whether "edge" is in "edge_list" and removes it.

    Parameters
    ----------
    edge_list : list or set
        List or set of edges.
    edge : tuple
        Edge.

    Returns
    -------
    edge_list : list or set
        Updated list or set of edges with "edge" removed if it was contained
        in "edge_list".

    """
    if edge in edge_list:
        edge_list.remove(edge)
    elif (edge[1], edge[0]) in edge_list:
        edge_list.remove((edge[1], edge[0]))
    return edge_list


# --- io utils ---
def open_tensorstore(path, driver):
    """
    Uploads segmentation mask stored as a directory of shard files.

    Parameters
    ----------
    path : str
        Path to directory containing shard files.
    driver : str
        Storage driver needed to read data at "path".

    Returns
    -------
    sparse_volume : dict
        Sparse image volume.

    """
    assert driver in SUPPORTED_DRIVERS, "Error! Driver is not supported!"
    arr = ts.open(
        {
            "driver": driver,
            "kvstore": {
                "driver": "gcs",
                "bucket": "allen-nd-goog",
                "path": path,
            },
            "context": {
                "cache_pool": {"total_bytes_limit": 1000000000},
                "cache_pool#remote": {"total_bytes_limit": 1000000000},
                "data_copy_concurrency": {"limit": 8},
            },
            "recheck_cached_data": "open",
        }
    ).result()
    if driver == "neuroglancer_precomputed":
        return arr[ts.d["channel"][0]]
    return arr


def read_tensorstore(path):
    """
    Reads neuroglancer_precomputed file at "path".

    Parameters
    ----------
    path : str
        Path to directory containing shard files.

    Returns
    -------
    ts.TensorStore
        Image volume.

    """
    dataset_ts = ts.open(
        {
            "driver": "neuroglancer_precomputed",
            "kvstore": {"driver": "file", "path": path},
        }
    ).result()
    return dataset_ts[ts.d["channel"][0]]


def read_tensorstore(arr, xyz, shape, from_center=True):
    chunk = get_chunk(arr, xyz, shape, from_center=from_center)
    return chunk.read().result()


def get_chunk(arr, xyz, shape, from_center=True):
    start, end = get_start_end(xyz, shape, from_center=from_center)
    return deepcopy(
        arr[start[0]: end[0], start[1]: end[1], start[2]: end[2]]
    )


def read_n5(path):
    """
    Reads n5 file at "path".

    Parameters
    ----------
    path : str
        Path to n5.

    Returns
    -------
    np.array
        Image volume.
    """
    return zarr.open(zarr.N5FSStore(path), "r").volume


def read_tif(path):
    """
    Reads tif file at "path".

    Parameters
    ----------
    path : str
        Path to tif.

    Returns
    -------
    np.array
        Image volume.
    """
    return imread(path)


def write_txt(path, contents):
    """
    Writes "contents" to a .txt file at "path".

    Parameters
    ----------
    path : str
        Path that .txt file is written to.
    contents : list[str]
        Contents to be written to .txt file.

    Returns
    -------
    None

    """
    with open(path, "w") as file:
        for line in contents:
            file.write(line + "\n")


def write_json(path, contents):
    """
    Writes "contents" to a .json file at "path".

    Parameters
    ----------
    path : str
        Path that .txt file is written to.
    contents : dict
        Contents to be written to .txt file.

    Returns
    -------
    None

    """
    with open(path, "w") as f:
        json.dump(contents, f)



"""
Created on Wed Jan 10 12:00:00 2023

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

import os

import numpy as np

from segmentation_skeleton_metrics import utils
from segmentation_skeleton_metrics.merge_metric import MergeMetric
from segmentation_skeleton_metrics.split_metric import SplitMetric


def run_evaluation(
    target_swc_dir,
    target_labels,
    pred_swc_dir,
    pred_labels,
    anisotropy=[1.0, 1.0, 1.0],
    filetype=None,
    log_dir=None,
    swc_log=False,
    txt_log=False,
):
    """
    Evaluates a predicted segmentation in terms of the number of splits
    and merges.

    Parameters
    ----------
    target_swc_dir : str
        Path to directory of swc files of the target segmentation.
    target_labels : np.array, ts.TensorStore, dict, or str
        Target segmentation mask or path to it.
    pred_swc_dir : str
        Path to directory of swc files of the predicted
        segmentation.
    pred_labels : np.array, ts.TensorStore, dict, or str
        Predicted segmentation mask or path to it.
    anisotropy : list[float], optional
        Image to real-world coordinates scaling factor for (x, y, z) which is
        applied to swc files.
    filetype : str, optional
        File type of target_labels and pred_labels if path is provided.
        The default is None.
    log_dir : str, optional
        Directory where logged information (i.e. swc log and txt log)
        is saved.
    swc_log : bool, optional
        Indicates whether to store swc files that indicates where mistakes are
        located.
    txt_log : bool, optional
        Indicates whether to store a txt files that contains (x, y, z)
        coordinates of where mistakes are located.

    Returns
    -------
    stats : dict
        Dictionary of (stat, value) pairs that were computed during the
        evaluation.

    """
    if log_dir is not None:
        utils.rmdir(log_dir)

    # Determine valid ids
    valid_ids = set()
    for f in utils.listdir(pred_swc_dir, ext=".swc"):
        f = int(f.replace(".swc", ""))
        valid_ids.add(f)

    # Split evaluation
    split_evaluator = SplitMetric(
        target_swc_dir,
        pred_labels,
        anisotropy=anisotropy,
        filetype=filetype,
        prefix="split-",
        log_dir=log_dir,
        swc_log=swc_log,
        txt_log=txt_log,
        valid_ids=valid_ids,
    )
    split_evaluator.detect_mistakes()

    # Merge evaluation
    merge_evaluator = MergeMetric(
        pred_swc_dir,
        target_labels,
        anisotropy=anisotropy,
        filetype=filetype,
        prefix="merge-",
        log_dir=log_dir,
        swc_log=swc_log,
        txt_log=txt_log,
    )

    # merge_evaluator.detect_mistakes()

    # Compute stats
    stats = dict()
    target_graphs = split_evaluator.graphs
    stats.update(compute_stats(split_evaluator, target_graphs, "split"))
    stats.update(compute_stats(merge_evaluator, target_graphs, "merge"))
    stats["num_mistakes"] = split_evaluator.site_cnt + merge_evaluator.site_cnt
    stats["edge_accuracy"] = compute_edge_accuracy(
        split_evaluator, merge_evaluator, target_graphs
    )

    utils.write_json(os.path.join(log_dir, "stats.json"), stats)
    return stats


def compute_stats(evaluator, list_of_graphs, x):
    """
    Computes various statistics that provide an evaluation of the
    predicted segmentation mask.

    Parameters
    ----------
    evaluator: MergeMetric or SplitMetric
        SegmentationMetric type object in which "detect_mistakes"
        has been run.
    list_of_graphs : list[networkx.Graph]
        List of graphs where each graph corresponds to a neuron.

    Returns
    -------
    stats : dict
        Dictionary where the keys are the names of statistics that were
        computed and the corresponding values are the numerical values were
        computed.

    """
    site_cnt = evaluator.site_cnt
    edge_cnt = evaluator.edge_cnt
    total_edges = count_edges(list_of_graphs)
    stats = {
        x + "_cnt": site_cnt,
        x + "_edge_cnt": edge_cnt,
        x + "_edge_ratio": edge_cnt / total_edges,
    }
    return stats


def compute_edge_accuracy(eval1, eval2, list_of_graphs):
    """
    Computes the percentage of correctly reconstructed edges from
    "list_of_graphs".

    Parameters
    ----------
    eval1 : MergeMetric or SplitMetric
        Type of SegmentationMetric.
    eval2 : MergeMetric or SplitMetric
        Type of SegmentationMetric.
    list_of_graphs : list[networkx.Graph]
        List of graphs where each graph corresponds to a neuron.

    Returns
    -------
    edge_accuracy : float
        Percentage of correctly reconstructed edges from
        "list_of_graphs".

    """
    e1 = eval1.edge_cnt
    e2 = eval2.edge_cnt
    total_edges = count_edges(list_of_graphs)
    return 1 - (e1 + e2) / total_edges


def count_edges(list_of_graphs):
    """
    Counts number of edges in 'list_of_graphs'.

    Parameters
    ----------
    graph : list[networkx.Graph]
        List of graphs where each graph corresponds to a neuron.

    Returns
    -------
    cnt : int
        Number of edges in 'list_of_graphs'.

    """
    return np.sum([graph.number_of_edges() for graph in list_of_graphs])



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

import os
from abc import ABC, abstractmethod

import numpy as np
import tensorstore as ts

from segmentation_skeleton_metrics import nx_utils, swc_utils, utils

SUPPORTED_FILETYPES = ["tif", "n5"]


class SegmentationMetrics(ABC):
    """
    Class that evaluates a segmentation in terms of the number of
    splits and merges.

    """

    def __init__(
        self,
        swc_path,
        labels,
        anisotropy=[1.0, 1.0, 1.0],
        filetype=None,
        prefix="",
        log_dir=None,
        swc_log=False,
        txt_log=False,
        valid_ids=None,
    ):
        """
        Constructs object that evaluates a predicted segmentation.

        Parameters
        ----------
        ...

        Returns
        -------
        None.

        """
        # Graph and labels
        self.anisotropy = anisotropy
        self.valid_ids = valid_ids
        print(swc_path)
        if os.path.isdir(swc_path):
            self.graphs = swc_utils.dir_to_graphs(swc_path, anisotropy=anisotropy)
        else:
            self.graphs = [swc_utils.file_to_graph(swc_path, anisotropy=anisotropy)]

        if type(labels) is str:
            self.labels = self.init_labels(labels, filetype)
        else:
            self.labels = labels

        # Mistake trackers
        self.site_cnt = 0
        self.edge_cnt = 0

        # Mistake logs
        self.prefix = prefix
        self.log_dir = log_dir
        self.swc_log = swc_log
        self.txt_log = txt_log
        if self.log_dir is not None:
            utils.mkdir(self.log_dir)
        if self.swc_log:
            self.swc_dir = os.path.join(self.log_dir, "swc_files")
            utils.mkdir(self.swc_dir)
        if self.txt_log:
            self.mistakes_log = ["# xyz1,  xyz2,  swc1,  swc2"]

    def init_labels(self, path, filetype):
        """
        Initializes a volume by uploading file with extension "filetype".

        Parameters
        ----------
        path : str
            Path to image volume.
        filetype : str
            Extension of file to be uploaded, supported values include tif, n5,
            and tensorstore.

        Returns
        -------
        dict
            Image volume.

        """
        assert filetype is not None, "Must provide filetype to upload image!"
        assert filetype in SUPPORTED_FILETYPES, "Filetype is not supported!"
        if filetype == "tensorstore":
            return utils.read_tensorstore(path)
        elif filetype == "n5":
            return utils.read_n5(path)
        else:
            return utils.read_tif(path)

    def get_labels(self):
        """
        Gets list of all unique labels in "self.labels".

        Parameters
        ----------
        None

        Returns
        -------
        list[int]
            List of all unique labels in "self.labels".

        """
        if type(self.labels) == np.array:
            labels = np.unique(self.labels)
        else:
            labels = np.unique(list(self.labels.keys()))
        return [label for label in labels if label != 0]

    def get_label(self, graph, i):
        """
        Gets label of node "i".

        Parameters
        ----------
        graph : networkx.Graph
            Graph which represents a neuron.
        i : int
            Node in "graph".

        Returns
        -------
        int
           Label of node "i".

        """
        return self._get_label(nx_utils.get_xyz(graph, i))

    def _get_label(self, xyz):
        """
        Gets label at image coordinates "xyz".

        Parameters
        ----------
        xyz : tuple[int]
            Coordinates that index into "self.labels".

        Returns
        -------
        int
           Label at image coordinates "xyz".

        """
        if type(self.labels) == dict:
            return 0 if xyz not in self.labels.keys() else self.labels[xyz]
        elif type(self.labels) == ts.TensorStore:
            return int(self.labels[xyz].read().result())
        else:
            return self.labels[xyz]

    def count_edges(self):
        """
        Counts number of edges in all graphs in "self.graphs".

        Parameters
        ----------
        None

        Returns
        -------
        int

        """
        total_edges = 0
        for graph in self.graphs:
            total_edges += graph.number_of_edges()
        return total_edges

    def is_mistake(self, a, b):
        """
        Checks if "a" and "b" are positive and not equal.

        Parameters
        ----------
        a : int
            label at node i.
        b : int
            label at node j.

        Returns
        -------
        bool
            Indicates whether there is a mistake.

        """
        if self.valid_ids is not None:
            if a not in self.valid_ids or b not in self.valid_ids:
                return False
        return (a != 0 and b != 0) and (a != b)

    def log(self, graph, edge_list):
        """
        Logs xyz coordinates of mistake and swc ids in a list if "txt_log"
        or writes an swc file if "log_swc".

        Parameters
        ----------
        graph : networkx.Graph
            Graph that contains edges in "edge_list".
        edge_list : list[tuple]
            Edges that correspond to a mistake.

        Returns
        -------
        None

        """
        if self.swc_log:
            self.swc_logger(graph, edge_list)

        if self.txt_log:
            self.txt_logger(graph, edge_list)

    def swc_logger(self, graph, edge_list):
        """
        Logs mistakes in an swc file.

        Parameters
        ----------
        graph : networkx.graph
            Graph that contains edges in "edge_list".
        edge_list : list[tuple]
            Edges that correspond to a mistake.

        Returns
        -------
        None

        """
        red = " 1.0 0.0 0.0"
        entries = swc_utils.make_entries(graph, edge_list, self.anisotropy)
        fn = self.prefix + str(self.site_cnt) + ".swc"
        path = os.path.join(self.swc_dir, fn)
        swc_utils.write_swc(path, entries, color=red)

    def txt_logger(self, graph, edge_list):
        """
        Logs xyz coordinates of mistake and swc ids in a list.

        Parameters
        ----------
        graph : networkx.graph
            Graph that contains edges in "edge_list".
        edge_list : list[tuple]
            Edges that correspond to a mistake.
        """
        for pair in edge_list:
            xyz_str = ""
            labels_str = ""
            for k in pair:
                xyz = swc_utils.node_to_world(graph, k, self.anisotropy)
                xyz_str += str(xyz) + ", "
                labels_str += str(self.get_label(graph, k)) + ", "
            self.mistakes_log.append(xyz_str + labels_str)

    def write_results(self, fn):
        """
        Writes "self.txt_log" to local machine at "self.log_dir".

        Parameters
        ----------
        fn : str
            Filename of text log.
        edge_list : list[tuple]
            List of edges that correspond to a mistakes.

        Returns
        -------
        None.

        """
        if self.txt_log:
            path = os.path.join(self.log_dir, fn + ".txt")
            utils.write_txt(path, self.mistakes_log)

    @abstractmethod
    def detect_mistakes(self):
        """
        Detects differences between corresponding labels of graph and volume.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        pass

    @abstractmethod
    def mistake_search(self):
        """
        Determines whether a complex mistake is a misalignment between the
        volume and graph or a true mistake.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        pass



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 15 12:00:00 2023

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

from random import sample


def sample_node(graph):
    """
    Samples a random node from "graph"

    Parameters
    ----------
    graph : networkx.Graph
        Graph in which node is sampled from.

    Returns
    -------
    int
        Node contained in graph.

    """
    return sample(list(graph.nodes), 1)[0]


def get_nbs(graph, i):
    """
    Gets neighbors of node "i".

    Parameters
    ----------
    graph : networkx.Graph
        Graph that represents a neuron.
    i : int
        Node of "graph".

    Returns
    -------
    list[int]
        List of neighbors of node "i".

    """
    return list(graph.neighbors(i))


def get_xyz(graph, i):
    """
    Gets (x,y,z) coordinates of node "i".

    Parameters
    ----------
    graph : networkx.Graph
        Graph that represents a neuron.
    i : int
        Node of "graph".

    Returns
    -------
    tuple
        The (x,y,z) coordinates of node "i".

    """
    return tuple(graph.nodes[i]["xyz"])



# -*- coding: utf-8 -*-
"""
Created on Wed June 5 16:00:00 2023

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

import os

import networkx as nx
import numpy as np

from segmentation_skeleton_metrics import nx_utils, utils


def make_entries(graph, edge_list, anisotropy):
    """
    Makes a list of entries to be written in an swc file.

    Parameters
    ----------
    graph : networkx.Graph
        Graph that edges in "edge_list" belong to.
    edge_list : list[tuple[int]]
        List of edges to be written to an swc file.
    anisotropy : list[float]
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files.

    Returns
    -------
    list[str]
        List of swc file entries to be written.

    """
    reindex = dict()
    for i, j in edge_list:
        if len(reindex) < 1:
            entry, reindex = make_entry(graph, i, -1, reindex, anisotropy)
            entry_list = [entry]
        entry, reindex = make_entry(graph, j, reindex[i], reindex, anisotropy)
        entry_list.append(entry)
    return entry_list


def make_entry(graph, i, parent, reindex, anisotropy):
    """
    Makes an entry to be written in an swc file.

    Parameters
    ----------
    graph : networkx.Graph
        Graph that "i" and "parent" belong to.
    i : int
        Node that entry corresponds to.
    parent : int
         Parent of node "i".
    anisotropy : list[float]
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files.

    """
    reindex[i] = len(reindex) + 1
    x, y, z = tuple(map(str, node_to_world(graph, i, anisotropy)))
    return [x, y, z, 8, parent], reindex


def node_to_world(graph, i, anisotropy):
    """
    Converts "xyz" from image coordinates to real-world coordinates.

    Parameters
    ----------
    i : int
        Node to be querried for (x, y, z) coordinates which are then converted
        to read-world coordinates.
    anisotropy : list[float]
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files.

    Returns
    -------
    list[float]
        Transformed coordinates.

    """
    return to_world(nx_utils.get_xyz(graph, i), anisotropy)


def to_world(xyz, anisotropy):
    """
    Converts "xyz" from image coordinates to real-world coordinates.

    Parameters
    ----------
    xyz : tuple or list
        Coordinates to be transformed.
    anisotropy : list[float]
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files.

    Returns
    -------
    list[float]
        Transformed coordinates.

    """
    return [xyz[i] * anisotropy[i] for i in range(3)]


def write_swc(path, entry_list, color=None):
    """
    Writes an swc file.

    Parameters
    ----------
    path : str
        Path on local machine that swc file will be written to.
    entry_list : list[list]
        List of entries that will be written to an swc file.
    color : str, optional
        Color of nodes. The default is None.

    Returns
    -------
    None.

    """
    with open(path, "w") as f:
        if color is not None:
            f.write("# COLOR" + color)
        else:
            f.write("# id, type, z, y, x, r, pid")
        f.write("\n")
        for i, entry in enumerate(entry_list):
            f.write(str(i + 1) + " " + str(0) + " ")
            for x in entry:
                f.write(str(x) + " ")
            f.write("\n")


def dir_to_graphs(swc_dir, anisotropy=[1.0, 1.0, 1.0]):
    """
    Converts directory of swc files to a list of graphs.

    Parameters
    ----------
    swc_dir : str
        Path to directory containing swc files.
    anisotropy : list[float], optional
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files.

    Returns
    -------
    list_of_graphs : list[networkx.Graph]
        List of graphs where each graph represents a neuron.

    """
    list_of_graphs = []
    for graph_id, f in enumerate(utils.listdir(swc_dir, ext=".swc")):
        path = os.path.join(swc_dir, f)
        graph = nx.Graph(file_name=f, graph_id=graph_id)
        graph = file_to_graph(path, graph=graph, anisotropy=anisotropy)
        list_of_graphs.append(graph)
    return list_of_graphs


def file_to_graph(path, graph=None, anisotropy=[1.0, 1.0, 1.0]):
    """
    Reads an swc file and constructs an undirected graph from it.

    Parameters
    ----------
    path : str
        Path to swc file to be read.
    graph : networkx.Graph
        Graph that info from swc file will be stored as a graph.
    anisotropy : list[float], optional
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files.

    Returns
    -------
    networkx.Graph
        Graph constructed from an swc file.

    """
    if graph is None:
        graph = nx.Graph()

    with open(path, "r") as f:
        offset = [0, 0, 0]
        for line in f.readlines():
            if line.startswith("# OFFSET"):
                parts = line.split()
                offset = read_xyz(parts[2:5])
            if not line.startswith("#") and len(line) > 0:
                parts = line.split()
                child = int(parts[0])
                parent = int(parts[-1])
                xyz = read_xyz(parts[2:5], anisotropy=anisotropy, offset=offset)
                graph.add_node(child, xyz=xyz, radius=parts[-2])
                if parent != -1:
                    graph.add_edge(parent, child)
    return graph


def read_xyz(xyz, anisotropy=[1.0, 1.0, 1.0], offset=[0, 0, 0]):
    """
    Reads the (x,y,z) coordinates from an swc file, then reverses and scales
    them.

    Parameters
    ----------
    xyz : str
        (x,y,z) coordinates.
    anisotropy : list[float], optional
        Image to real-world coordinates scaling factors applied to "xyz".
    offset : list[int], optional
        Offset of (x, y, z) coordinates in swc file.

    Returns
    -------
    tuple
        The (x,y,z) coordinates of an entry from an swc file in real-world
        coordinates.

    """
    xyz = [int(np.round(float(xyz[i]) / anisotropy[i] + offset[i])) for i in range(3)]
    return tuple(xyz)
